#include <iostream>
using namespace std;

const int fixedStringLength = 4;


void printArray(const int *A,int n);
void printAll(char s[],int currentIndex,int stringLength);
void printAllC(char s[], int currentIndex,int stringLength,int maxWeight);
void printAllPermutations(int currentPerm[],int unavailable[],int currentIndex, int permLength);
void storeAll(char s[],int currentIndex,char M[][fixedStringLength+1], int& counter);

void testPrintAll();
void testPrintAllC();
void testPrintAllPermutations();
void testStoreAll();

//To design a recursive algorithm, we assume we have a function that solves a problem
//of the same kind, but of smaller size, and then we try to use this function to help us
//solve the larger problem. As you might have noticed from previous programming assignments,
//this process bears a strong similarity to mathematical induction. We will try to illustrate
//this point of view using this exercise. 

void main()
{
	testPrintAll();
	testStoreAll();
	testPrintAllC();
	testPrintAllPermutations();	
}

void testPrintAll(){
	char s[100];
	printAll(s,0,6);
}

void testStoreAll(){
	char s[fixedStringLength+1];
	char M[10000][fixedStringLength+1];
	int counter = 0;
	storeAll(s,0,M,counter);
	for(int i=0;i<counter;i++) cout << M[i] << endl;
}
void testPrintAllC(){
	char s[2000];
	// the prototype of printAllC is :
	//void printAllC(char s[], int currentIndex,int stringLength,int curWeight,int maxWeight)
	printAllC(s,0,8,2);
}
void testPrintAllPermutations(){
	int av[4]={0};
	int A[4];
	printAllPermutations(A,av,0,4);
}
//As mentioned previously, when we are trying to design a recursive algorithm, we assume that
//we already have a function that works on problems of smaller size. However, on this problem,
//this approach does not seem to work at first sight : even if one has a function ,say, that prints
//all abc strings of length 3, that doesn't make it easier to make a function that prints
//all abc strings of length 4. So we have to define our subproblems in a way that makes it easy
//for us to solve the original problem. That's why, instead of defining a subproblem as printing
//abc strings of smaller sizes, we define a subproblem to consist in printing all abc-strings of 
//full size stringLength *such that their first n characters are fixed*. (currentIndex represents the
//number of fixed characters,which can also be considered as the index of the first non-fixed array entry, 
//because arrays are 0-based).
//So, let's assume we have a function that lets us print all abc strings of length stringLength having their 
//first n+1 characters fixed. Now,it is easy to use this function to print all abc strings of length stringLength 
//having their first n characters fixed : just put 'a' at the nth position and call the function that
//leaves the first n+1 characters fixed, then put 'b' at the nth position and call the function again,
//and then put 'c' at the nth position and call the function again. So we now have our recursive formulation.
//The base case of the function is extremely easy to understand: When all the array is fixed, we only have
//to print it.
//This is  what we're doing in the printAll function.
void printAll(char s[],int currentIndex,int stringLength){
	if(currentIndex==stringLength){
		s[currentIndex]='\0';
		cout << s << endl;
		return;
	}
	s[currentIndex]='a';
	printAll(s,currentIndex+1,stringLength);
	s[currentIndex]='b';
	printAll(s,currentIndex+1,stringLength);
	s[currentIndex]='c';
	printAll(s,currentIndex+1,stringLength);
}
//We removed the argument stringLength, since the length is assumed to be 4
//storeAll is a very simple modification of the previous problem: instead of printing
//to the screen, it prints to an array.
void storeAll(char s[],int currentIndex,char M[][fixedStringLength+1], int& counter){
	if(currentIndex==fixedStringLength){
		for(int i=0;i<fixedStringLength;i++)
			M[counter][i]=s[i];
		M[counter][fixedStringLength]='\0';
		counter++;
		return;
	}
	s[currentIndex]='a';
	storeAll(s,currentIndex+1,M,counter);
	s[currentIndex]='b';
	storeAll(s,currentIndex+1,M,counter);
	s[currentIndex]='c';
	storeAll(s,currentIndex+1,M,counter);
	
}
//printAllC is a bit more difficult.
//The function printAllC prints all binary strings of length stringLength, having their first n
//characters fixed, and containing not more than maxWeight '1' characters between index n and 
//index stringLength included. With this general idea in mind, it is relatively easy to give a sketch of
//a proof of correctness for printAllC, by induction on stringLength-n:
//for a value of stringLength-n=0, all the characters are fixed , and this function is obviously
//correct. Suppose now this function is correct for stringLength-n = z (n>0). Let's prove it correct
//stringLength-n=z+1; the function first sets the nth element to '0' and calls itself recursively
//on a larger n,and therefore a smaller value of stringLength -n. But we assumed it was correct for smaller values
//of stringLength-n, and therefore it will print correctly all output strings having a '0' in the nth
//place. 
//after having taken care of the case where the nth character is a '0' ,we now consider the case
//where the nth character is a '1'. First we check if ones are at all allowed, then we set the nth
//character to '1' , we freeze the n+1 first characters, and we call the function recursively, without
//forgetting to decrease the number of allowable '1' characters by one. Since  the function will
//work correctly on this subproblem (by induction), the function will work correctly when the
//nth character is set to '1'. Therefore printAllC will always work correctly.
void printAllC(char s[], int n,int stringLength,int maxWeight){
	if(n==stringLength){
		s[n]='\0';
		cout << s << endl;
		return;
	}
	s[n]='0';
	printAllC(s,n+1,stringLength,maxWeight);
	if(maxWeight>0) {
		s[n]='1';
		printAllC(s,n+1,stringLength,maxWeight-1);
	}

}
void printArray(const int *A,int n)
{
	for (int i=0;i<n;i++) 
		cout<<A[i]<<" ";
	cout<<endl;
}

//printAllPermutations will print all permutations of length permLength which first currentIndex 
//positions are fixed. Since a given number can appear only once in a given permutation,
//the numbers have already been used are flagged in the array unavailable. Of course, this
//information is redundant, it is only there for efficiency purposes.
void printAllPermutations(int currentPerm[],int unavailable[],int currentIndex, int permLength){
	if(currentIndex==permLength){
		printArray(currentPerm,permLength);
		return;
	}
	for(int i=1;i<=permLength;i++){
		if(unavailable[i-1]==0){
			currentPerm[currentIndex]=i;
			unavailable[i-1]=1;
			printAllPermutations(currentPerm,unavailable,currentIndex+1, permLength);
			unavailable[i-1]=0;
		}
	}
}
